/* vim:sw=4:sts=4
 *
 * Library to use the Gtk2 widget library from Lua 5.0
 * Copyright (C) 2005 Wolfgang Oertl
 * Use this software under the terms of the GPLv2.
 *
 * Revision history:
 *  2005-07-24	first public release
 */

#include <lua.h>
#include <lauxlib.h>
#include <stdlib.h>		/* bsearch */
#include <string.h>		/* strcmp */
#include <stdio.h>		/* fopen & co */

#include <gtk/gtk.h>
#include "hash.h"
#include "structinfo.h"
/* #include <glib-object.h> */

/* settings */
#define MAX_DLL			5
#define MAX_CALLBACK_ARGS	6
#define MAX_FUNC_ARGS		20
#undef DEBUG

static const char *dll_list[] = {
#ifdef LINUX
    "/usr/lib/libgtk-x11-2.0.so",
    "/usr/lib/libgdk-x11-2.0.so",
    "/usr/lib/libpango-1.0.so",
#endif
#ifdef WIN32
    "C:/GTK/bin/libgtk-win32-2.0-0.dll",
    "C:/GTK/bin/libgdk-win32-2.0-0.dll",
    "C:/GTK/bin/libpango-1.0-0.dll",
    "libgobject-2.0-0.dll",
#endif
    NULL
};


/* globals */
static lua_State *L_global = NULL;
static int dl_count = 0;
// static int gvalue_ofs;

/* redirect common functions to save on dynamic relocation data */
/* this can save memory of maybe 0.5 kB - not very effective... */
static void lua_pushstring_i(lua_State *L, const char *s) { lua_pushstring(L, s); }
#define lua_pushstring lua_pushstring_i
static void lua_settop_i(lua_State *L, int i) { lua_settop(L, i); }
#define lua_settop lua_settop_i
static void lua_pushvalue_i(lua_State *L, int i) { lua_pushvalue(L, i); }
#define lua_pushvalue lua_pushvalue_i
static int lua_gettop_i(lua_State *L) { return lua_gettop(L); }
#define lua_gettop lua_gettop_i
static void lua_rawset_i(lua_State *L, int i) { lua_rawset(L, i); }
#define lua_rawset lua_rawset_i
static const void *lua_topointer_i(lua_State *L, int i) { return lua_topointer(L, i); }
#define lua_topointer lua_topointer_i
static int printf_i(const char *fmt, ...) { va_list va; va_start(va, fmt);
 return vprintf(fmt, va); }
#define printf printf_i


#ifdef LINUX
#include <dlfcn.h>
#include <ffi.h>		/* foreign function interface library */
#define EXPORT
#define DLLOOKUP dlsym(dl_handle[i], func_name)
#define DLLOAD(name) dlopen(name, RTLD_LAZY)
static void *dl_handle[MAX_DLL];
#endif

#ifdef WIN32
#include <windows.h>
#include "ffi.h"
#define EXPORT __declspec(dllexport)
#define DLLOOKUP (FARPROC) GetProcAddress(dl_handle[i], func_name)
#define DLLOAD(name) LoadLibrary(name)
static HINSTANCE dl_handle[MAX_DLL];
#endif

/* data structures */

/* one such structure per connected callback */
struct callback_info {
    int id;
    GSignalQuery query;
};
/* query: signal_id, signal_name, itype, signal_flags, return_type, n_params,
 * param_types */

union gtk_arg_types {
    void	*p;
    long	l;
    double	d;
    float	f;
};

/**
 * Info about a pending callback.  When the callback happens, I don't have
 * the "L" (the Lua thread structure) available, so I have to keep it in C's
 * space.
 */
struct callback {
    void *widget;
    struct callback_info *cbi;
    union gtk_arg_types data[MAX_CALLBACK_ARGS];
};

/* information about a function in the shared library */
struct func_info {
    void *func;
    const char *name;		/* full name in dynamic library;
				   usually malloc()ed */
    const unsigned char *args_info;
};

/**
 * entry (type "userdata") in the meta table of a widget.  These entries are
 * created on the first access to a method or attribute to make later
 * uses quicker.
 */
struct meta_entry {
    int	type;				/* 0=function, >0 = attribute type */
    union {
	struct func_info fi;		/* 12 bytes */
	const struct struct_elem *se;   /* 4 bytes */
    };
    char name[0];
};

/* functions from the DLLs */

static int (*g_type_from_name_ptr)(const char *class_name);
static int (*g_type_parent_ptr)(int class_nr);
static const char *(*g_type_name_ptr)(int class_nr);
static void (*g_type_query_ptr)(int type_nr, GTypeQuery *query);
static void (*g_object_unref_ptr)(gpointer object);
static void (*gtk_object_sink_ptr)(GtkObject *object);
static gpointer (*g_object_ref_ptr)(gpointer object);
static GType *(*g_type_interfaces_ptr)(GType type, guint *n_interfaces);
static int (*g_signal_lookup_ptr)(const char *name, GType type);
static void (*gtk_object_destroy_ptr)(GtkObject *object);
static void (*g_signal_query_ptr)(int signal_id, GSignalQuery *query);
static int (*g_signal_connect_data_ptr)(gpointer instance, const char *signame,
    GCallback c_handler, gpointer data, GClosureNotify destroy_data,
    GConnectFlags connect_flags);
static GtkType (*gtk_object_get_type_ptr)(void);
static gboolean (*g_type_check_instance_is_a_ptr)(GTypeInstance *instance,
    GType iface_type);
static GType (*g_type_fundamental_ptr)(GType type_id);
static void (*g_free_ptr)(void*);


/**
 * A few functions from the dynamic libraries are used from this code.  To
 * avoid linking the DLL directly, they are looked up once at init time after
 * manually loading the DLL.
 */
struct linkfunc {
    const char *name;
    void **ptr;
};

#define LINK(s) { #s, (void (*)) &s##_ptr },
static const struct linkfunc dl_link[] = {
    LINK(g_type_from_name)
#define g_type_from_name g_type_from_name_ptr
    LINK(g_type_parent)
#define g_type_parent g_type_parent_ptr
    LINK(g_type_name)
#define g_type_name g_type_name_ptr
    LINK(g_type_query)
#define g_type_query g_type_query_ptr
    LINK(g_type_interfaces)
#define g_type_interfaces g_type_interfaces_ptr
    LINK(g_type_fundamental)
#define g_type_fundamental g_type_fundamental_ptr
    LINK(g_object_unref)
#define g_object_unref g_object_unref_ptr
    LINK(g_object_ref)
#define g_object_ref g_object_ref_ptr
    LINK(g_signal_lookup)
#define g_signal_lookup g_signal_lookup_ptr
    LINK(g_signal_query)
#define g_signal_query g_signal_query_ptr
    LINK(g_signal_connect_data)
#define g_signal_connect_data g_signal_connect_data_ptr
    LINK(g_free)
#define g_free g_free_ptr
    LINK(g_type_check_instance_is_a)
#define g_type_check_instance_is_a g_type_check_instance_is_a_ptr
    LINK(gtk_object_destroy)
#define gtk_object_destroy gtk_object_destroy_ptr
    LINK(gtk_object_sink)
#define gtk_object_sink gtk_object_sink_ptr
    LINK(gtk_object_get_type)
#define gtk_object_get_type gtk_object_get_type_ptr
    { NULL, NULL }
};


/* external tables with hash tables */
extern struct hash_info funclist_hash;
extern struct hash_info enumlist_hash;

extern char override_data[];
extern int override_size;


/* function declarations */
static int gtk_index(lua_State *L);
static int l_gtk_connect(lua_State *L);


/* debug */

#ifdef DEBUG
/* only for debugging */
#include "../../lua-5.0.2/src/lstate.h"

void dump_stack(lua_State *L)
{
    StkId ptr;
    char s[100];

    const char *lua_type_names[] = {
	"nil", "bool", "lightuserdata", "number",
	"string", "table", "function", "userdata", "thread" };

    printf("STACK DUMP\n");
    for(ptr=L->base; ptr<L->top; ptr++) {
	switch (ptr->tt) {
	    case LUA_TBOOLEAN:
		sprintf(s, "%s", ptr->value.b ? "true" : "false");
		break;
	    case LUA_TUSERDATA:
	    case LUA_TLIGHTUSERDATA:
	    case LUA_TFUNCTION:
	    case LUA_TTABLE:
		sprintf(s, "%p", ptr->value.p);
		break;
	    case LUA_TSTRING:
		sprintf(s, "%s", ((char*)&ptr->value.gc->ts.tsv) + sizeof(TString));
		break;
	    case LUA_TNUMBER:
		sprintf(s, "%f", ptr->value.n);
		break;
	    default:
		strcpy(s, "?");
	}
	printf("  %d %s %s\n",
	    ptr - L->base + 1,
	    lua_type_names[ptr->tt],
	    s);
    }
}
#endif


/**
 * The list of types is included directly and not compiled separately, because
 * it needs to include ffi.h, which is dependent on the operating system,
 * and I want to keep the #ifs and #elses to a minimum.
 */
#include "_typelist.c"

#define NAME(ofs) (struct_strings + (ofs))

/* comparison for binary search */
static int struct_compare(const void *key, const void *item)
{
    const char *name = NAME(((struct struct_info*)item)->name_ofs);
    return strcmp(key, name);
}


/**
 * Structures are primarily looked up by their number, but in order to resolve
 * the hierarchy a lookup by name is also required.  A hash table would be too
 * difficult to use in this case.
 */
static inline struct struct_info *find_struct(const char *class_name)
{
    struct struct_info *p;
    p = (struct struct_info*) bsearch(class_name, struct_list,
	struct_count, sizeof(*struct_list), struct_compare);
    return p;
}


/**
 * Look for the function in the dynamic library.  If it is not found, this is
 * not an error, because many tries for different widget types may be
 * necessary to find one method.
 * Returns 0 if the function hasn't been found, 1 otherwise.
 */
static int find_func(const char *func_name, struct func_info *fi)
{
    int i;

    /* Find the function in the hash table.  Note that when using
     * minimal hash tables (see documentation), a function may be found
     * even if it doesn't exist.
     */
    if (!(fi->args_info = hash_search(func_name, &funclist_hash)))
	return 0;

    /* find the function in the shared library */
    for (i=0; i<dl_count; i++)
	if ((fi->func = DLLOOKUP)) {
	    fi->name = func_name; // strdup(func_name);
	    return 1;
	}

    /* only an error for full hashes */
    if (funclist_hash.full_hash)
	printf("ERROR: function %s not found in shared library.\n",
	    func_name);
    return 0;
}


/**
 * Look for an attribute of the given class.  The attributes are ordered
 * by their offset within the structure, but not strictly - unions have
 * their attributes at the same offset.
 */
static const struct struct_elem *find_attribute(struct struct_info *si,
    const char *attr_name)
{
    const struct struct_elem *e, *e_end;

    e = elem_list + si->elem_start;

    for (e_end = e + si->elem_count; e < e_end; e++)
	if (!strcmp(attr_name, NAME(e->name_ofs)))
	    return e;

    return NULL;
}


/**
 * Called for a structure, return a printable representation.
 * stack: widget
 */
static int gtk_tostring(lua_State *L)
{
    char buf[50];
    void **p = (void**) lua_topointer(L, 1);

    lua_getmetatable(L, -1);
    lua_pushstring(L, "_classname");
    lua_rawget(L, -2);

    sprintf(buf, "%s at %p", lua_tostring(L, -1), *p);
    lua_pushstring(L, buf);

    return 1;
}


/**
 * Handler for the __gc event on GtkObjects.  It is wise to call
 * gtk_object_unref in this case.
 *
 * XXX gtk_object_destroy is not always correct.  Sometimes g_object_unref
 * should be used instead.
 */
static int gobject_collect(struct lua_State *L)
{
    luaL_checktype(L, 1, LUA_TUSERDATA);
    GtkObject **obj = (GtkObject**) lua_topointer(L, 1);

#if 0
    lua_getmetatable(L, -1);
    lua_pushstring(L, "_classname");
    lua_rawget(L, -2);
    const char *s = lua_tostring(L, -1);
    printf("unref %s at %p: %d\n", s, *obj,
	(*obj)->parent_instance.ref_count);
    lua_pop(L, 2);
#endif

    if ((*obj)->parent_instance.ref_count > 0)
	g_object_unref(*obj);
    else
	printf("WARNING: try to unref a freed object!\n");

    /* to avoid freeing this reference again */
    *obj = NULL;
    return 0;
}


/**
 * Given the structure nr, retrieve or create the metaclass for this type of
 * widget.  If the given widget type has a base class, recurse to make that,
 * too.
 *
 * Stack input: nothing
 * Stack output: one value: the metaclass.
 */
static int get_widget_meta(lua_State *L, int struct_nr, int recurse, void *p)
{
    const struct struct_info *sptr = struct_list + struct_nr, *sptr2;

    lua_getglobal(L, "_meta_tables");
    lua_pushstring(L, NAME(sptr->name_ofs));
    lua_rawget(L, -2);

    /* stack: _meta_tables meta|nil -- if it exists, return it. */
    if (!lua_isnil(L, -1)) {
	lua_remove(L, -2);
	return 1;
    }

    /* create a new meta table with __index, _classname, _struct, _parent */
    lua_pop(L, 1);
    /* stack: _meta_tables */
    lua_newtable(L);

    /* store in _meta_tables */
    lua_pushstring(L, NAME(sptr->name_ofs));
    lua_pushvalue(L, -2);
    lua_rawset(L, -4);
    lua_remove(L, -2);

    /* stack: meta -- add the __index function */
    lua_pushstring(L, "__index");
    lua_pushcfunction(L, gtk_index);
    lua_rawset(L, -3);

    lua_pushstring(L, "__tostring");
    lua_pushcfunction(L, gtk_tostring);
    lua_rawset(L, -3);

    lua_pushstring(L, "_struct");
    lua_pushlightuserdata(L, (void*) sptr);
    lua_rawset(L, -3);

    lua_pushstring(L, "_classname");
    lua_pushstring(L, NAME(sptr->name_ofs));
    lua_rawset(L, -3);

    /* determine GTK type number */
    GTypeQuery query;
    int type_nr;
    const char *parent_name;
    query.type_name = NULL;

    type_nr = g_type_from_name(NAME(sptr->name_ofs));
    if (type_nr == 0) {
	printf("WARNING: the type %s is not registered.\n", NAME(sptr->name_ofs));
	return 1;
    }

    lua_pushstring(L, "_gtktype");
    lua_pushnumber(L, type_nr);
    lua_rawset(L, -3);

    /* a destructor for GtkObjects and their derived types.  XXX hack! */
    if (p && G_TYPE_IS_OBJECT(type_nr)
	    && !memcmp(NAME(sptr->name_ofs), "Gtk", 3)
	    && GTK_IS_OBJECT(p)) {
	// printf("destuctor for type %s\n", NAME(sptr->name_ofs));
	lua_pushstring(L, "__gc");
	lua_pushcfunction(L, gobject_collect);
	lua_rawset(L, -3);
    }

    if (!recurse)
	return 1;

    /* determine the name of the parent class, if any */
    type_nr = g_type_parent(type_nr);
    if (type_nr == 0) {
	/* actually not a problem - the base classes have no parent */
	/* printf("WARNING: no parent found for %s\n", sptr->name); */
	return 1;
    }

    parent_name = g_type_name(type_nr);
    if (!parent_name) {
	printf("ERROR: Gtk has no name for type nr. %d\n", type_nr);
	return 0;
    }

    sptr2 = find_struct(parent_name);
    if (!sptr2) {
	// printf("WARNING: unknown base structure %s of %s\n", parent_name,
	//     sptr->name);
	return 1;
    }

    struct_nr = sptr2 - struct_list;
    get_widget_meta(L, struct_nr, 1, p);

    /* add _parent -- used by gtk_index */
    lua_pushstring(L, "_parent");
    lua_insert(L, -2);
    lua_rawset(L, -3);
    return 1;
}


/**
 * Push a new widget (userdata) onto the stack.  The only difference between
 * make_widget and make_struct is the parameter "is_widget" to get_widget_meta;
 * it controls whether an attempt is made to find parent structures.
 *
 * You can either specify the structure number, or the class_name.
 */
static int make_widget(lua_State *L, void *p, int struct_nr,
    const char *class_name)
{
    void **widget;

    if (class_name) {
	struct struct_info *p;
	p = find_struct(class_name);
	struct_nr = p - struct_list;
    }

    /* with meta table */
    widget = (void**) lua_newuserdata(L, sizeof(void*));
    *widget = p;
    get_widget_meta(L, struct_nr, 1, p);
    lua_setmetatable(L, -2);

    /* a new reference is now held */
    {
	const char *name = NAME(struct_list[struct_nr].name_ofs);
	// XXX this is too simple a test.  Some GtkXXX objects are NOT
	// derived from GtkObject.  If not, don't do this.
	// EXCEPT: g_object_ref should be called for an existing
	// object; if it is returned from a xxx_new function, don't call it
	// because it already has ref count 1 upon creation.
	if (!memcmp(name, "Gtk", 3) && GTK_IS_OBJECT(p)) {
	    // int type = GTK_OBJECT_TYPE(p);
	    // printf("make_widget: %s is type %d\n", name, type);
	    g_object_ref(p);
	    gtk_object_sink(p);
	    // printf("ref/sink %s %d\n", name, ((GObject*)p)->ref_count);
	}
    }

    /*
    printf("after make widget for %s: refcount is %d\n",
	NAME(struct_list[struct_nr].name_ofs), ((GObject*)p)->ref_count);
    */

    return 1;
}


static int make_struct(lua_State *L, void *p, int struct_nr)
{
    void **st;

    /* with meta table */
    st = (void**) lua_newuserdata(L, sizeof(void*));
    *st = p;
    get_widget_meta(L, struct_nr, 0, p);
    lua_setmetatable(L, -2);
    return 1;
}


/**
 * When using object:connect(), g_object_connect would be called with
 * unwanted effects (SEGV).
 * The only difference to l_gtk_connect is _gtk at stack position #1.
 */
int gtk_object_connect_override(lua_State *L)
{
    lua_getglobal(L, "_gtk");
    lua_insert(L, 1);
    return l_gtk_connect(L);
}


/**
 * A parameter for a callback must be pushed onto the stack.  The type to
 * use depends on the "type" (from the g_query_signal results).
 */
static void push_a_value(lua_State *L, GType type, union gtk_arg_types *data)
{
    if (G_TYPE_IS_FUNDAMENTAL(type)) {
	switch (type) {
	    case G_TYPE_BOOLEAN:
		lua_pushboolean(L, data->l);
		break;
	    case G_TYPE_INT:
	    case G_TYPE_UINT:
	    case G_TYPE_LONG:
	    case G_TYPE_INT64:
	    case G_TYPE_UINT64:
	    case G_TYPE_ENUM:
		lua_pushnumber(L, data->l);
		break;
	    case G_TYPE_FLOAT:
		lua_pushnumber(L, data->f);
		break;
	    case G_TYPE_DOUBLE:
		lua_pushnumber(L, data->d);
		break;
	    case G_TYPE_POINTER:
		// some opaque structure -- no way to access that data.
		// printf("what type is this arg?\n");
		lua_pushlightuserdata(L, data->p);
		break;
	    case G_TYPE_STRING:
		lua_pushstring(L, data->p);
		break;
	    // XXX handle more fundamental types
	    case G_TYPE_NONE:
		printf("strange... an argument of type NONE?\n");
		break;
	    default:
		printf("Warning: unhandled fundamental type %ld\n", type >> 2);
		lua_pushnumber(L, data->l);
	}
    } else {
	/* not a fundamental type */
	const char *name = g_type_name(type);
	if (!name) {
	    printf("ERROR callback argument is not a valid type!\n");
	    lua_pushnumber(L, data->l);
	} else {
	    make_widget(L, data->p, -1, name);
	}
    }
}


/**
 * This is the most important function in this module.  Call a function from
 * the dynamically loaded library, using the information about parameters
 * compiled in from automatically generated configuration files.
 *
 * Stack: parameters starting at "index".
 */
static int do_call(lua_State *L, struct func_info *fi, int index)
{
    union gtk_arg_types ffi_args[MAX_FUNC_ARGS];
    ffi_type *argtypes[MAX_FUNC_ARGS];
    void *argvalues[MAX_FUNC_ARGS];
    ffi_cif cif;

    /* output variables */
    union gtk_arg_types retval, retvals[MAX_FUNC_ARGS];

    const unsigned char *s, *s_end;
    int arg_struct_nr;
    int ret_struct_nr=0;
    const struct ffi_type_map_t *arg_type, *ret_type=NULL;
    int i, type;
    int stack_top = lua_gettop(L);

    /* DEBUG */
#if 0
    s = fi->args_info;
    s_end = s + *s;
    s++;
    for (i = -1; s <= s_end; i++) {
	arg_type = &ffi_type_map[*s - '0'];
	s ++;
	if (arg_type->with_struct) {
	    arg_struct_nr = * (unsigned short*) s;
	    s += 2;
	} else {
	    arg_struct_nr = 0;
	}
	if (i == -1) {
	    printf("- %s %s(", arg_type->name, fi->name);
	} else {
	    printf("%s%s", i == 0 ? "" : ", ", arg_type->name);
	}
    }
    printf(")\n");
    // dump_stack(L);
#endif
    /* END DEBUG */

    /*
     * ffi_args	    array of values passed to the function
     * argtypes	    array of types for ffi_call
     * argvalues    array of pointers to the values
     */

    /* build the call stack by parsing the parameter list */
    s = fi->args_info;
    s_end = s + *s;
    s++;

    for (i = -1; s <= s_end; i++) {
	arg_type = &ffi_type_map[*s - '0'];
	s ++;
	if (arg_type->with_struct) {
	    arg_struct_nr = * (unsigned short*) s;
	    s += 2;
	} else {
	    arg_struct_nr = -1;
	}

	/* the first "argument" is actually the return value */
	if (i == -1) {
	    ret_type = arg_type;
	    ret_struct_nr = arg_struct_nr;
	    continue;
	}

	if (index+i > stack_top) {
	    printf("WARNING: function %s expects more arguments -> nil\n",
		fi->name);
	    // return 0;
	    type = LUA_TNIL;
	} else 
	    type = lua_type(L, index+i);

	argtypes[i] = arg_type->type;
	argvalues[i] = &ffi_args[i].l;

	switch (arg_type->at) {
	    case AT_BOOL:
		luaL_checktype(L, index+i, LUA_TBOOLEAN);
		ffi_args[i].l = (long) lua_toboolean(L, index+i);
		break;
	    case AT_LONG:
		ffi_args[i].l = (long) luaL_checknumber(L, index+i);
		break;
	    case AT_DOUBLE:
		ffi_args[i].d = (double) luaL_checknumber(L, index+i);
		break;
	    case AT_STRING:
		ffi_args[i].p = (void*) luaL_checkstring(L, index+i);
		break;
	    case AT_WIDGET:
		if (type == LUA_TNIL)
		    ffi_args[i].p = NULL;
		else if (type == LUA_TUSERDATA) {

		// XXX check type -- using some glib function!
#if 0
		// check type
		if (!lua_getmetatable(L, index+i)) {
		    printf("ERROR: widget has no meta table.\n");
		} else {
		    lua_pushstring(L, "_classname");
		    lua_rawget(L, -2);
		    const char *class_name = lua_tostring(L, -1);
		    lua_pop(L, 2);
			printf("verify type %s is %s\n", class_name,
			    struct_list[arg_struct_nr].name);
		    }
#endif
		    ffi_args[i].p = * (void**) lua_topointer(L, index+i);
		} else {
		    printf("ERROR: incompatible argument #%d for %s (type %d)\n",
			i, fi->name, type);
		    return 0;
		}
		break;

	    case AT_STRUCTPTR:
		if (type == LUA_TNIL) {
		    const struct struct_info *si = struct_list + arg_struct_nr;
		    if (!strcmp(NAME(si->name_ofs), "GValue")) {
			GValue *gvalue = (GValue*) alloca(sizeof *gvalue);
			gvalue->g_type = 0;
			ffi_args[i].p = gvalue;
			break;
		    }
		}
		/* fall through */

	    case AT_POINTER:
		if (type == LUA_TNIL)
		    ffi_args[i].p = NULL;
		else if (type == LUA_TLIGHTUSERDATA)
		    ffi_args[i].p = (void*) lua_topointer(L, index+i);
		else if (type == LUA_TUSERDATA)
		    ffi_args[i].p = * (void**) lua_topointer(L, index+i);
		else if (type == LUA_TSTRING)
		    ffi_args[i].p = (void*) lua_tostring(L, index+i);
		else if (arg_type->at == AT_STRUCTPTR) {
		    const struct struct_info *si = struct_list + arg_struct_nr;
		    if (strcmp(NAME(si->name_ofs), "GValue")) {
			printf("ERROR: incompatible argument #%d to %s: "
			    "target=%s\n", i, fi->name, NAME(si->name_ofs));
			return 0;
		    }
		    // try to make this structure
		    GValue *gvalue = (GValue*) alloca(sizeof *gvalue);
		    switch (type) {
			case LUA_TNUMBER:
			gvalue->g_type = G_TYPE_UINT;
			gvalue->data[0].v_uint = lua_tonumber(L, index+i);
			break;

			case LUA_TBOOLEAN:
			gvalue->g_type = G_TYPE_BOOLEAN;
			gvalue->data[0].v_uint = lua_toboolean(L, index+i) ? 1: 0;
			break;

			case LUA_TSTRING:
			gvalue->g_type = G_TYPE_POINTER;
			gvalue->data[0].v_pointer = (void*) lua_tostring(L,
			    index+i);
			break;

			default:
			printf("ERROR: can't coerce type %d to GValue\n",
			    type);
			return 0;
		    }

		    ffi_args[i].p = (void*) gvalue;
		} else {
		    printf("ERROR: don't know what to do.\n");
		    return 0;
		}
		break;

	    case AT_LONGPTR:
		/* Can be input: e.g. gdk_pango_layout_get_clip_region
		 * Use as such when an array (of numbers) is given.
		 */
		if (lua_istable(L, index+i)) {
		    int j, n, *ar;
		    n = luaL_getn(L, index+i);
		    ar = (int*) alloca(sizeof(*ar) * n);
		    for (j=0; j<n; j++) {
			lua_rawgeti(L, index+i, j+1);
			ar[j] = lua_tonumber(L, -1);
			lua_pop(L, 1);
		    }
		    ffi_args[i].p = (void*) ar;
		    break;
		}

		/* This is an OUTPUT value - initialize with whatever the
		 * user passed as parameter. */
		ffi_args[i].p = (void*) &retvals[i];
		retvals[i].l = lua_tonumber(L, index+i);
		break;

	    case AT_VARARG:
		/* use the rest of the stack */
		for (; index+i <= stack_top; i++) {
		    int type = lua_type(L, index+i);
		    switch (type) {
			case LUA_TNUMBER:
			    argtypes[i] = &ffi_type_uint;
			    argvalues[i] = &ffi_args[i].l;
			    ffi_args[i].l = (long) lua_tonumber(L, index+i);
			    break;
			case LUA_TSTRING:
			    argtypes[i] = &ffi_type_pointer;
			    argvalues[i] = &ffi_args[i].p;
			    ffi_args[i].p = (void*) lua_tostring(L, index+i);
			    break;
			case LUA_TNIL:
			    argtypes[i] = &ffi_type_pointer;
			    argvalues[i] = &ffi_args[i].p;
			    ffi_args[i].p = NULL;
			    break;
			default:
			    printf("Unhandled vararg type %d of function %s\n",
				type, fi->name);
		    }
		}
		/* counter the i++ of next iteration. */
		i--;
		break;

	    // XXX some types are still missing, like double

	    default:
		printf("ERROR: %s: arg %d (type %s) not handled\n",
		    fi->name,
		    i+1,
		    arg_type->name);
		ffi_args[i].l = 0;
	}
    }

    /* any arguments left over? (note: more might be used than given -> nil) */
    if (index + i - 1 < stack_top) {
	printf("Warning: superfluous arguments to %s\n", fi->name);
    }


    /* call the function */
    if (ffi_prep_cif(&cif, FFI_DEFAULT_ABI, i,
	ret_type->type,
	argtypes) != FFI_OK) {
	printf("FFI call to %s couldn't be initialized\n", fi->name);
	return 0;
    }

    ffi_call(&cif, fi->func, &retval, argvalues);

    int stack_pos = lua_gettop(L);

    /* evaluate the return value */
    // printf("- %s: return type is %s\n", fi->name, ret_type->name);
    switch (ret_type->at) {
	case AT_LONG:
	    lua_pushnumber(L, retval.l);
	    break;
	case AT_BOOL:
	    lua_pushboolean(L, retval.l);
	    break;
	case AT_WIDGET:
	    make_widget(L, retval.p, ret_struct_nr, NULL);
	    break;
	case AT_STRUCTPTR:
	    make_struct(L, retval.p, ret_struct_nr);
	    // lua_pushlightuserdata(L, retval.p);
	    break;
	case AT_VOID:
	    break;
	case AT_STRING:
	case AT_POINTER:	// XXX maybe not always OK?
	    lua_pushstring(L, (char*) retval.p);
	    break;
	default:
	    printf("ERROR: unhandled return type %s of %s\n",
		ret_type->name, fi->name);
    }

    /* Return output parameters, too.  Unfortunately, this requires
     * another pass at parsing the parameters.
     */
    for (s = fi->args_info + 1, i = -1; s <= s_end; i++) {
	arg_type = &ffi_type_map[*s - '0'];
	s ++;
	if (arg_type->with_struct) {
	    arg_struct_nr = * (unsigned short*) s;
	    s += 2;
	} else {
	    arg_struct_nr = 0;
	}
	if (i == -1)
	    continue;

	switch (arg_type->at) {
		case AT_LONGPTR:
		    lua_pushnumber(L, retvals[i].l);
		    break;
		case AT_STRUCTPTR:;
		    const struct struct_info *si = struct_list + arg_struct_nr;
		    if (!strcmp(NAME(si->name_ofs), "GValue")) {
			GValue *gvalue = (GValue*) ffi_args[i].p;
			push_a_value(L, gvalue->g_type,
			    (union gtk_arg_types*) &gvalue->data);
		    }
		    break;
		// XXX other AT_xxxPTR things are missing.
		default:
		    break;
	}
    }

    /* return number of return values now on the stack. */
    return lua_gettop(L) - stack_pos;
}


/**
 * Call any gtk function through this catch-all API.
 * Stack: 1=name of the function, 2 and up=arguments to the function
 */
static int l_call(lua_State *L)
{
    const char *func_name = luaL_checkstring(L, 1);
    struct func_info fi;

    if (find_func(func_name, &fi)) {
	int rc = do_call(L, &fi, 2);
	// free(fi.name);
	return rc;
    }

    printf("ERROR: function %s not found\n", func_name);
    return 0;
}


/**
 * Allocate a structure, initialize with zero and return.
 */
static int l_new(lua_State *L)
{
    const char *struct_name = luaL_checkstring(L, 1);
    struct struct_info *si;

    if (!(si=find_struct(struct_name))) {
	printf("ERROR: structure %s not found\n", struct_name);
	return 0;
    }

    /* allocate structure plus a pointer to it just before it */
    /* I guess (hope) that this takes care of freeing it later */
    char *p = (char*) lua_newuserdata(L, si->struct_size + sizeof(void*));
    * (void**) p = p + sizeof(void*);
    memset(p+sizeof(void*), 0, si->struct_size);

    /* XXX note: g_object_unref shouldn't be called on this... */
    /* So this shouldn't be a GObject! */
    get_widget_meta(L, si - struct_list, 0, NULL);
    lua_setmetatable(L, -2);

    return 1;
}


/**
 * After a method has been looked up, this function is called to do the
 * invoction of the corresponding gtk function.
 */
static int l_call_func(lua_State *L)
{
    struct meta_entry *me = (struct meta_entry*) lua_topointer(L,
	lua_upvalueindex(1));
    return do_call(L, &me->fi, 1);
}


/**
 * Given a pointer, a bit offset and a bit length, retrieve the value.
 */
static long get_bits(const unsigned char *ptr, int bitofs, int bitlen)
{
    unsigned int val;

    if (bitlen == 0 || bitlen > 32) {
	printf("Error: access to attribute of size %d not supported\n",
	    bitlen);
	return 0;
    }

    ptr += bitofs / 8;
    bitofs = bitofs % 8;
    unsigned int mask = 0xffffffff;
    if (bitlen < 32) {
	mask = (1 << bitlen) - 1;
    }

    // XXX not tested for bitofs > 0
    val = ((* (unsigned int*) ptr) >> bitofs) & mask;
    return val;
}


/**
 * A meta entry is on the top of the stack; use it to retrieve the method
 * pointer or attribute value.
 * stack: 1=widget, 2=key, 3=dest metatable, 4=current metatable,... meta entry
 */
static int gtk_handle_entry(lua_State *L)
{
    /* an override -- just return it */
    if (lua_iscfunction(L, -1) || lua_isfunction(L, -1))
	return 1;

    const struct meta_entry *me = lua_topointer(L, -1);
    const struct ffi_type_map_t *arg_type;
    unsigned char *ptr;
    long int v;

    if (me->type == 0) {
	lua_pushlightuserdata(L, (void*) me);
	lua_pushcclosure(L, l_call_func, 1);
	return 1;
    }

    /* handle attribute access */
    ptr = * (unsigned char**) lua_topointer(L, 1);
    arg_type = &ffi_type_map[me->type - '0'];
    switch (arg_type->at) {
	case AT_LONG:
	    v = get_bits(ptr, me->se->bit_offset, me->se->bit_length);
	    lua_pushnumber(L, v);
	    return 1;

	case AT_WIDGET:
	    v = get_bits(ptr, me->se->bit_offset, me->se->bit_length);
	    make_widget(L, (void*) v, me->se->type_detail, NULL);
	    return 1;

	case AT_STRUCT:
	    make_struct(L, ptr + me->se->bit_offset/8, me->se->type_detail);
	    return 1;

	// XXX handle more data types!
	default:
	    printf("ERROR: unhandled attribute type %s (attribute %s)\n",
		arg_type->name, me->name);
    }
    return 0;
}


/**
 * Given a class name in the form "GtkVBox" and a method name like
 * "pack_start", construct a function name like "gtk_vbox_pack_start".
 */
static void make_func_name(char *buf, const char *class_name,
    const char *attr_name)
{
    const char *s = class_name;
    char *out = buf;

    if (!strncmp(s, "Pango", 5)) {
	strcpy(out, "gdk");
	out += 3;
    }

    while (*s) {
	if (*s >= 'A' && *s <= 'Z') {
	    if (out != buf)
		*out++ = '_';
	    *out++ = *s + ('a' - 'A');
	} else {
	    *out++ = *s;
	}
	s++;
    }
    *out++ = '_';
    strcpy(out, attr_name);
}


/**
 * A method has been found and should now be called.
 * input stack: parameters to the function
 * upvalues: the name of the function - func - args_info
 */
static int l_gtk_call_wrapper(lua_State *L)
{
    struct func_info fi;
    fi.name = (char*) lua_tostring(L, lua_upvalueindex(1));
    fi.func = (void*) lua_topointer(L, lua_upvalueindex(2));
    fi.args_info = lua_topointer(L, lua_upvalueindex(3));
    return do_call(L, &fi, 1);
}


/**
 * Handle accesses of "gtk.xxx", where xxx may be any gtk function, used mainly
 * for gtk.xxx_new(), and ENUMs.
 *
 * input stack: 1=table, 2=value
 * output: either an integer (for ENUMs) or a closure (for functions).
 */
static int l_gtk_lookup(lua_State *L)
{
    const char *s = lua_tostring(L, -1);
    struct func_info fi;
    char func_name[50];

    /* if it starts with an uppercase letter, it's probably an ENUM. */
    if (s[0] >= 'A' && s[0] <= 'Z') {
	const unsigned char *res = hash_search(s, &enumlist_hash);
	if (res) {
	    // stored value has 0-4 bytes.
	    int len = (int) *res;
	    int val = 0;
	    while (len--)
		val = (val << 8) + *++res;
	    lua_pushnumber(L, val);
	    return 1;
	}
    }

    sprintf(func_name, "gtk_%s", s);
    if (!find_func(func_name, &fi)) {
	printf("ERROR: attribute or method not found: %s\n", s);
	return 0;
    }

    /* A function has been found, so return a closure that can call it. */
    lua_pushstring(L, func_name);
    lua_pushlightuserdata(L, fi.func);
    lua_pushlightuserdata(L, (void*) fi.args_info);
    lua_pushcclosure(L, l_gtk_call_wrapper, 3);
    return 1;
}


/**
 * __index function for the metatable used for userdata (widgets); it must
 * handle accesses to methods and attributes found in this class or any base
 * class.  Once the method or attribute has been found, it is inserted
 * into the widget's table to avoid looking it up again.
 *
 * Stack: 1=widget, 2=key
 */
static int gtk_index(lua_State *L)
{
    char name[50];
    struct func_info fi;
    int recursed = 0;
    const char *class_name, *attr_name;
    attr_name = luaL_checkstring(L, 2);

    if (!lua_getmetatable(L, 1)) {
	printf("ERROR: no meta table of widget in gtk_index on access "
	    "to attribute or method %s\n", attr_name);
	return 0;
    }
    lua_pushvalue(L, -1);

    /* stack: 1=widget, 2=key, 3=destination metatable, 4=current metatable */

again:
    /* retrieve the class name, an element of the meta table */
    lua_pushstring(L, "_classname");
    lua_rawget(L, 4);
    class_name = lua_tostring(L, -1);
    lua_pop(L, 1);

    /* already in this metatable? */
    lua_pushvalue(L, 2);
    lua_rawget(L, 4);
    if (!lua_isnil(L, -1)) {
	if (recursed) {
	    /* add to base table */
	    lua_pushvalue(L, 2);
	    lua_pushvalue(L, 5);
	    lua_rawset(L, 3);
	}
	return gtk_handle_entry(L);
    }
    lua_pop(L, 1);	/* get rid of nil */
	

    /* not found -- maybe a (not yet mapped) function? */
    make_func_name(name, class_name, attr_name);

    /* Maybe a function loaded from override.lua?  Those are globals */
    lua_getglobal(L, name);
    if (!lua_isnil(L, -1))
	goto is_override;
    lua_pop(L, 1);

    /* maybe with override? */
    lua_getglobal(L, "_gtk");
    lua_pushstring(L, name);
    lua_rawget(L, -2);
    if (!lua_isnil(L, -1)) {
is_override:
	// printf("OVERRIDE FOR %s\n", name);
	lua_pushvalue(L, 2);
	lua_pushvalue(L, -2);
	lua_rawset(L, 3);
	// dump_stack(L);
	return 1;
    }
    lua_pop(L, 2);

    if (find_func(name, &fi))
	goto found_func;
    

    

    /* maybe a gdk_ function for GdkXXX objects? */
    if (!recursed && !strncmp(class_name, "Gdk", 3)) {
	sprintf(name, "gdk_%s", attr_name);
	if (find_func(name, &fi))
	    goto found_func;
    }

    /* maybe it is an attribute? */
    const struct struct_elem *se = NULL;
    lua_pushstring(L, "_struct");
    lua_rawget(L, 4);
    if (!lua_isnil(L, -1)) {
	struct struct_info *si = (struct struct_info*) lua_topointer(L, -1);
	se = find_attribute(si, attr_name);
    }
    lua_pop(L, 1);

    if (se) {
	struct meta_entry *me = (struct meta_entry*) lua_newuserdata(L,
	    sizeof(*me));
	me->type = se->type;
	me->se = se;
	lua_pushvalue(L, 2);
	lua_pushvalue(L, -2);
	lua_rawset(L, 3);
	return gtk_handle_entry(L);
    }

    /* retrieve the class type, an element of the meta table */
    lua_pushstring(L, "_gtktype");
    lua_rawget(L, 4);
    int gtk_type = lua_tonumber(L, -1);
    lua_pop(L, 1);

    /* the object may implement one or more interfaces -- check them */
    guint n_interfaces;
    int i;
    GType *gtypes = g_type_interfaces(gtk_type, &n_interfaces);

    for (i=0; i<n_interfaces; i++) {
	class_name = g_type_name(gtypes[i]);
	/* not found -- maybe a (not yet mapped) function? */
	make_func_name(name, class_name, attr_name);

	/* XXX duplicated code from above -- check for overrides */
	lua_getglobal(L, name);
	if (!lua_isnil(L, -1)) {
	    g_free(gtypes);
	    goto is_override;
	}
	lua_pop(L, 1);

	if (find_func(name, &fi)) {
	    g_free(gtypes);
	    goto found_func;
	}
    }
    g_free(gtypes);

    /* maybe there is a base class? */
    lua_pushstring(L, "_parent");
    lua_rawget(L, 4);
    if (!lua_isnil(L, 5)) {
	lua_remove(L, 4);
	recursed = 1;
	goto again;
    }

    /* give up */
    printf("!! NOT FOUND: %s\n", attr_name);
    return 0;

found_func:;
    struct meta_entry *me = (struct meta_entry*) lua_newuserdata(L,
	sizeof(*me) + strlen(name) + 1);
    me->type = 0;
    memcpy(&me->fi, &fi, sizeof(me->fi));
    strcpy(me->name, name);
    lua_pushvalue(L, 2);
    lua_pushvalue(L, -2);
    lua_rawset(L, 3);
    return gtk_handle_entry(L);
}



/**
 * Gtk calls a signal handler; do this now.
 * Note: this runs in the lua_State that the init function was called with.
 * Note 2: no assumptions can be made about the contents of the stack, as
 *  this is a callback it may be invoked any time (just my guess to be on the
 *  safe side).
 *
 * Returns: a value to return to Gtk.
 */
static int do_callback(lua_State *L, struct callback *cb)
{
    int i, n, val, return_count, stack_top, t, m=0;
    GType type;
    GTypeQuery query;

    /* The Lua function to call and their (arbitrary, user supplied) parameters
     * are Lua data, so they have to be stored within the Lua VM.  Therefore
     * the struct callback only has an ID, which is a key to the global _gtk
     * table that maps to a value: either just the Lua function, or a table
     * with the function followed by the parameters.
     *
     * XXX could use a separate table like _gtk_callbacks
     */

    stack_top = lua_gettop(L);
    lua_getglobal(L, "_gtk");
    lua_pushnumber(L, cb->cbi->id);
    lua_rawget(L, -2);
    lua_remove(L, -2);

    /* stack: cbinfo */
    // stack_top = lua_gettop(L);
    t = lua_type(L, -1);

    /* callback with user parameters is stored as table */
    if (t == LUA_TTABLE) {
	lua_pushnil(L);
	/* table - cbinfo - nil */
	while (lua_next(L, stack_top+1) != 0) {
	    /* key at -2, value at -1 */
	    lua_insert(L, -2);
	    m ++;		/* count extra arguments */
	}
	lua_remove(L, stack_top+1);	/* eliminate table */
	m --;			/* don't count function */
	/* stack: table - func - arg... */
    } else if (t != LUA_TFUNCTION) {
	printf("ERROR - callback param is type %d!\n", t);
	return 0;
    }


    /* build a proper Lua widget from the GtkWidget pointer */
    type = G_TYPE_FROM_INSTANCE(cb->widget);
    g_type_query(type, &query);

    if (query.type == 0) {
	cb->widget = NULL;
	printf("ERROR: callback widget is not a valid type!\n");
	return 0;
    }
    make_widget(L, cb->widget, -1, query.type_name);


    /* push the arguments provided by Gtk for the signal handler */
    n = cb->cbi->query.n_params;
    for (i=0; i<n; i++) {
	GType type = cb->cbi->query.param_types[i]
	    & ~G_SIGNAL_TYPE_STATIC_SCOPE;
	push_a_value(L, type, &cb->data[i]);
    }

    /* determine whether a return value is expected */
    GType return_type = cb->cbi->query.return_type
	& ~G_SIGNAL_TYPE_STATIC_SCOPE;
    return_count = (return_type == G_TYPE_NONE) ? 0 : 1;

    lua_call(L, n+m+1, return_count);

    /* Retrieve the return value (if any).  Zero is returned if no return value
     * is expected by Gtk.
     */
    val = 0;

    switch (return_type) {
	case G_TYPE_NONE:
	    break;
	case G_TYPE_BOOLEAN:
	    val = lua_toboolean(L, -1);
	    break;
	// XXX handle more return types!
	default:
	    printf("ERROR: unhandled callback return type %ld\n", return_type);
    }


    /* make sure the stack is back to the original state */
    lua_settop(L, stack_top);

    return val;
}


/* Callbacks with different number of parameters.  This is called directly
 * by Gtk; the _last_ parameter is the "data" that was passed to
 * g_signal_connect_data.
 *
 * NOTE: sometimes a return value is expected!  Look at callback info for
 * more information.
 */
static int g_callback_0(void *widget, struct callback_info *cbi)
{
    struct callback cb1;
    cb1.widget = widget;
    cb1.cbi = cbi;
    return do_callback(L_global, &cb1);
}

static int g_callback_1(void *widget, int data0, struct callback_info *cbi)
{
    struct callback cb1;
    cb1.widget = widget;
    cb1.cbi = cbi;
    cb1.data[0].l = data0;
    return do_callback(L_global, &cb1);
}

static int g_callback_2(void *widget, int data0, int data1, struct callback_info *cbi)
{
    struct callback cb1;
    cb1.widget = widget;
    cb1.cbi = cbi;
    cb1.data[0].l = data0;
    cb1.data[1].l = data1;
    return do_callback(L_global, &cb1);
}

static int g_callback_3(void *widget, int data0, int data1, int data2, struct callback_info *cbi)
{
    struct callback cb1;
    cb1.widget = widget;
    cb1.cbi = cbi;
    cb1.data[0].l = data0;
    cb1.data[1].l = data1;
    cb1.data[2].l = data2;
    return do_callback(L_global, &cb1);
}

static void (*g_callbacks[]) = {
    g_callback_0,
    g_callback_1,
    g_callback_2,
    g_callback_3
};


/**
 * Connect a signal to a Lua function.
 *
 * input: 1=api (table returned by init), 2=widget, 3=signal name,
 * 4=lua function, 5... extra parameters
 */
static int l_gtk_connect(lua_State *L)
{
    luaL_checktype(L, 1, LUA_TTABLE);
    luaL_checktype(L, 2, LUA_TUSERDATA);
    luaL_checktype(L, 3, LUA_TSTRING);
    luaL_checktype(L, 4, LUA_TFUNCTION);

    void *widget = * (void**) lua_topointer(L, 2);
    // XXX move into the table at stack pos 1?
    static int next_id = 1;
    struct callback_info *cb_info;
    GType type;
    guint signal_id;

    if (!widget) {
	printf("ERROR: trying to connect a NULL widget\n");
	return 0;
    }

    /* no extra args -> 4 */
    int stack_top = lua_gettop(L);

    type = G_OBJECT_TYPE(widget);
    const char *signame = luaL_checkstring(L, 3);
    if (!(signal_id = g_signal_lookup(signame, type))) {
	printf("ERROR: cannot find signal %s\n", signame);
	return 0;
    }

    /* g_signal_query(signal_id, &query); */
    cb_info = (struct callback_info*) malloc(sizeof *cb_info);
    g_signal_query(signal_id, &cb_info->query);

    /* stack: table - widget - name - func - .... */

    /* make an entry in the callback table: key=id, value=lua function
     * or value=table( 1=function, 2=arg1, 3=arg2, ...)
     */
    lua_pushnumber(L, next_id);
    if (stack_top == 4) {
	lua_pushvalue(L, 4);
    } else {
	lua_newtable(L);
	lua_pushnumber(L, 1);
	lua_pushvalue(L, 4);
	lua_rawset(L, -3);
	int i;
	for (i=5; i<=stack_top; i++) {
	    lua_pushnumber(L, i-3);
	    lua_pushvalue(L, i);
	    lua_rawset(L, -3);
	}
	/* stack: table - widget - name - func - ... - next_id - table */
    }
    lua_rawset(L, 1);
    lua_settop(L, 3);

    /* stack: table - widget - name */
    if (cb_info->query.n_params >= sizeof(g_callbacks) /
	sizeof(g_callbacks[0])) {
	printf("ERROR: can't handle callback with %d parameters.\n",
	    cb_info->query.n_params);
	return 0;
    }

    cb_info->id = next_id ++;
    g_signal_connect_data(widget, signame,
	g_callbacks[cb_info->query.n_params], cb_info, NULL, 0);

    return 0;
}


/**
 * If you want you can call gtk.close() when finished with Gtk; it
 * currently only frees the DLLs, but could call a function that
 * frees all allocated Gtk memory -- unfortunately I haven't found it.
 */
static int l_close(lua_State *L)
{
    int i;

    for (i=0; i<dl_count; i++) {
	dlclose(dl_handle[i]);
	dl_handle[i] = NULL;
    }
    dl_count = 0;
    return 0;
}


/* methods directly callable from Lua; most go through __index */
static const luaL_reg methods[] = {
    {"__index",		l_gtk_lookup },
    {"call",		l_call },
    {"connect",		l_gtk_connect },
    {"close",		l_close },
    {"new",		l_new },
    {"gtk_object_connect", gtk_object_connect_override },
    { NULL, NULL }
};

/* initialize the library, returns a table */
EXPORT int luaopen_gtk2(lua_State *L)
{
    const char **dlname;
    const struct linkfunc *dl;
    struct func_info fi;

    /* load dynamic libraries */
    for (dlname=dll_list, dl_count=0; *dlname; dlname++) {
	if (!(dl_handle[dl_count] = DLLOAD(*dlname))) {
	    printf("Can't load dynamic library %s\n", *dlname);
	    return 0;
	}
	dl_count ++;
    }

    /* make a few functions easily accessible */
    for (dl=dl_link; dl->name; dl++) {
	if (!find_func(dl->name, &fi)) {
	    printf("ERROR: function %s not found in dynamic library.\n",
		dl->name);
	    return 0;
	}
	*(dl->ptr) = fi.func;
	// free(fi.name);
    }

    L_global = L;

    /* global variables */
    lua_pushstring(L, "_meta_tables");
    lua_newtable(L);
    lua_settable(L, LUA_GLOBALSINDEX);

    /* make the table to return, and make it global at _gtk */
    lua_newtable(L);
    lua_pushstring(L, "_gtk");
    lua_pushvalue(L, -2);
    lua_settable(L, LUA_GLOBALSINDEX);

    /* execute the glue library (compiled in) */
    lua_dobuffer(L, override_data, override_size, "override.lua");

    /* define meta table and functions */
    lua_pushvalue(L, -1);
    lua_setmetatable(L, -2);
    luaL_openlib(L, NULL, methods, 0);

    return 1;		/* one retval on the stack: api */
}

